/*
Copyright (C) 2011 by Blazing Games Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com.blazinggames.cgl.atlas
{
	/**
	 * This holds informatoin about an Image Atlas animation. These sequences
	 * consist of a series of frames, with the frames being clips from the
	 * image atlas.
	 * 
	 * @author Billy D. Spelchan
	 */	
	public class AnimationInfo
	{
// ***** VARIABLES *****
		
		private var _sequenceName:String;
		private var _frames:Array;
		
// ***** CONSTRUCTOR *****
		
		/** 
		 * Creates an empty unnamed sequence.
		 */		
		public function AnimationInfo()
		{
			_frames = new Array();
		}
		
// ***** METHODS - XML Handling *****
		
		/**
		 * Frame tags are children of the Animation tag and have the format 
		 * <FRAME number='frame' clip='clip id' x='offset x' y='offset y' duration='duration' /> 
		 *  
		 */
		public function setFromXML(xml:XML, atlas:ImageAtlas):void
		{
			var frames:XMLList = xml.elements("FRAME");
			
			_sequenceName = xml.@name;
			var numFrames:int = frames.length();
			var frameInfo:FrameInfo;
			for (var cntr:int = 0; cntr < numFrames; ++cntr) {
				frameInfo = new FrameInfo();
				frameInfo.setFromXML(xml.FRAME[cntr], atlas);
				addFrame(frameInfo, true);
			}
		}
		
		/**
		 * Creates a String that holds an XML tag for definining this 
		 * animation in XML. This can be useful for debugging and for tools 
		 * that generate animations as the then the animation can be created
		 * by manually adding frames and then converted into the appropriate
		 * ImageAtlas XML tags for storing the file.
		 * 
		 * @return String with the XML definition for the animation.
		 */		
		public function toXMLString():String
		{
			var s:String = "<ANIMATION name=\'"+_sequenceName+"\' >\n";
			for (var cntr:int = 0; cntr < _frames.length; ++cntr) {
				s = s + _frames[cntr].toXMLString() + "\n";				
			}
			s = s + "</ANIMATION>";
			
			return s;
		}

// ***** METHODS - Getters and Setters *****
		
		/**
		 * Sequence Names are strings that identify the animation and should 
		 * be unique, at least unique to the image atlas the sequence is in.
		 * 
		 * @return name of the animation sequence.
		 */		
		public function get sequenceName():String
		{
			return _sequenceName;
		}

		/**
		 * Sets the sequence name. Sequence names identify particular 
		 * animations so should be unique.
		 * 
		 * @param value name of the sequence.
		 */		
		public function set sequenceName(value:String):void
		{
			_sequenceName = value;
		}

		/**
		 * The number of frames that make up this animation sequence. 
		 * 
		 * @return The number of frames that make up this animation sequence. 
		 */		
		public function get length():int
		{
			return _frames.length;
		}

// ***** METHODS - Frame Management *****

		/**
		 * Adds a frame to the sequence. Default behavior is to append frame to
		 * end but can choose to keep the existing frame number.
		 *  
		 * @param frame the frame to add
		 * @param keepNumber if false, appends to end of clip otherwise uses frameNumber stored in frame.
		 * @return frame number frame was set to.
		 * 
		 */		
		public function addFrame(frame:FrameInfo, keepNumber:Boolean = false) : int
		{
			var frameNumber:int = _frames.length;
			if (keepNumber) 
				frameNumber = frame.frameNumber;
			_frames[frameNumber] = frame;
			frame.frameNumber = frameNumber;
			return frameNumber;
		}
		
		/**
		 * Returns the frame info for the indicated frame or null if no such 
		 * frame. 
		 * 
		 * @param frameNumber The frame number that information is desired for.
		 * @return FrameInfo for desired frame or null if no such frame.
		 */		
		public function getFrame(frameNumber:int):FrameInfo
		{
			return _frames[frameNumber];
		}

	}
}