/*
Copyright (C) 2011 by Blazing Games Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com.blazinggames.cgl.atlas
{
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.MovieClip;
	import flash.display.Scene;
	import flash.events.Event;
	import flash.utils.getTimer;
	
	/**
	 * Plays Image Atlas Animations using the MovieClip API.
	 * 
	 * @author Billy D. Spelchan 
	 */	
	public class AtlasMovie 
	extends MovieClip
	{
// ***** CONSTANTS *****
		
		public static const ANIMATION_FINISHED_EVENT:String = "AnimDone";
		
		public static const PLAYBACK_MANUAL:int = 0;
		public static const PLAYBACK_FORWARD:int = 1;
		public static const PLAYBACK_BACKWARD:int = 2;
		public static const PLAYBACK_YOYO:int = 3;
		
		private static const DEFAULT_ERROR_FRAME_WIDTH:int = 32;
		private static const DEFAULT_ERROR_FRAME_HEIGHT:int = 32;
		private static const DEFAULT_ERROR_FRAME_ARGB:uint = 0x80FF80FF;

// ***** VARIABLES *****
		
		private var _atlas:ImageAtlas;
		private var _bitmap:Bitmap;
		private var _frame:FrameInfo;
		private var _anim:AnimationInfo;
		private var _currentFrame:int;
		private var _currentDirection:int;
		private var _playbackType:int;
		private var _skipFrames:Boolean;
		private var _frameStartTime:int;
		private var _repeatCount:int;
		private var _currentRepetition:int;
		private var _isPlaying:Boolean;
		// error frame should be visible so developer knows problem exists
		private var _errorFrameWidth:int;
		private var _errorFrameHeight:int;
		private var _errorFrameColor:uint;
		private var _errorBitmapData:BitmapData;
		
// ***** CONSTRUCTOR *****
		
		/**
		 * Creates a movie clip that plays an Image Atlas animation. The 
		 * animation being played can be changed at any time using the
		 * setSequence command.
		 * 
		 * @param atlas The Image Atlas that holds the Animation Sequence(s) to be played.
		 * @param sequenceName The name of the animation sequence to be played
		 * @param playback The type of playback (see the PLAYBACK constants)
		 * @param skipFrames should skipping of frames be allowed.
		 * @param repeat Number of times to repeat the sequence.
		 */		
		public function AtlasMovie(	atlas:ImageAtlas, 
								   	sequenceName:String, 
								   	playback:int = PLAYBACK_FORWARD,
									skipFrames:Boolean = false,
									repeat:int = 0x07FFFFFF)
		{
			_atlas = atlas;
			_bitmap = new Bitmap();
			addChild(_bitmap);
			setErrorFrame();
			addEventListener(Event.ENTER_FRAME, onEnterFrame);
			setSequence(sequenceName, playback, skipFrames, repeat);
			super();
		}

// ***** Methods - setup and atlas manipulation *****

		/**
		 * Changes the animation sequence being played
		 * 
		 * @param sequenceName The name of the animation sequence to be played
		 * @param playback The type of playback (see the PLAYBACK constants)
		 * @param skipFrames should skipping of frames be allowed.
		 * @param repeat Number of times to repeat the sequence.
		 */		
		public function setSequence(sequenceName:String, 
									playback:int = PLAYBACK_FORWARD,
									skipFrames:Boolean = false,
									repeat:int = 0x07FFFFFF):void
		{
			_anim = _atlas.getAnimation(sequenceName);
			_playbackType = playback;
			_skipFrames = skipFrames;
			_repeatCount = repeat;
			_currentRepetition = 0;
			if (playback == PLAYBACK_BACKWARD) {
				_currentDirection = -1;
				if (_anim == null)
					showFrame(0);
				else
					showFrame(_anim.length - 1);
			} else {
				_currentDirection = 1;
				showFrame(0);
			}
			_isPlaying = true;
		}

		/**
		 * Defines the image to display if an invalid sequence is provided. 
		 * Having something visible when there is a problem with the sequence
		 * is preferable to nothing/logging/crashing as things will still run
		 * and there is a visible indication that there is a problem somewhere.
		 * 
		 * @param w width of the error frame
		 * @param h height of the error frame
		 * @param c color of the error frame (ideally transparent but up to you)
		 */		
		public function setErrorFrame(w:int = DEFAULT_ERROR_FRAME_WIDTH, 
									  h:int = DEFAULT_ERROR_FRAME_HEIGHT,
									  c:uint = DEFAULT_ERROR_FRAME_ARGB):void
		{
			_errorFrameWidth = w;
			_errorFrameHeight = h;
			_errorFrameColor = c;		
		}

		/**
		 * Shows the indicated frame. 
		 * 
		 * @param frame Frame number to be shown
		 */		
		public function showFrame(frame:int) :void
		{
			var validFrame:Boolean = true;
			_currentFrame = frame;
			if (_anim == null) {
				validFrame = false;
			} else {
				_frame = _anim.getFrame(_currentFrame);
				if (_frame != null) {
					if (_frame.clip != null) {
						_bitmap.bitmapData = _frame.clip.bitmapData;
						_bitmap.x = 0 - _frame.clip.origin.x + _frame.offset.x;
						_bitmap.y = 0 - _frame.clip.origin.y + _frame.offset.y;
						_bitmap.visible = true;
					} else
						validFrame = false;
				} else
					validFrame = false;
			}
			if ( ! validFrame ) {
				if (_errorBitmapData == null)
					_errorBitmapData = new BitmapData(_errorFrameWidth, _errorFrameHeight, true, _errorFrameColor);
				_bitmap.bitmapData = _errorBitmapData;
			}
			_frameStartTime = getTimer();
		}

		/**
		 * Internal routine for handing the animation timing. Am considering 
		 * switching to use of intervals for frame timing so this may be 
		 * depreciated.
		 * 
		 * @param evt the ENTER_FRAME event
		 */		
		protected function onEnterFrame(evt:Event) : void
		{
			if ( _anim == null) {
				nextFrame();
			}
			if  ( ! _isPlaying)  {
				return;
			}
			// see if time for next frame
			var frameTime:int = getTimer() - _frameStartTime;
			if (_frame == null) {
				frameTime = 0;
				nextFrame();
			} else if (frameTime >= _frame.duration) {
				if (_skipFrames) {
					while (frameTime > 0) {
						frameTime -= _frame.duration;
						nextFrame();
						if (_frame == null) frameTime = 0;
					}
					_frameStartTime -= frameTime;
				} else {
					nextFrame();
				}
			}
		}
		
// ***** Methods - From MovieClip *****
		
		/**
		 * Returns the number of the frame that is being displayed. 
		 * @return the current frame.
		 */		
		override public function get currentFrame():int
		{
			return _currentFrame;
		}

		/**
		 * In Image Atlas Animations, frames all share the name of the sequence.
		 * @return name of the current sequence 
		 */		
		override public function get currentFrameLabel():String
		{
			if (_anim == null)
				return "";
			else
				return _anim.sequenceName;
		}

		/**
		 * Goes to the indicated frame and starts playing the animation.
		 * Image Atlas uses sequences so scene name is treated as changing 
		 * the sequence.
		 * 
		 * @param frame Frame to start playing animation on
		 * @param scene Animation Sequence to be played if different.
		 */		
		override public function gotoAndPlay(frame:Object, scene:String=null):void
		{
			if (scene != null)
				setSequence(scene, _playbackType, _skipFrames, _repeatCount);
			var f:String = frame.toString();
			showFrame(parseInt(f));
			_isPlaying = true;
		}
		
		/**
		 * Goes to the indicated frame and STOPS playing the animation.
		 * Image Atlas uses sequences so scene name is treated as changing 
		 * the sequence.
		 * 
		 * @param frame Frame to start playing animation on
		 * @param scene Animation Sequence to be played if different.
		 */		
		override public function gotoAndStop(frame:Object, scene:String=null):void
		{
			if (scene != null)
				setSequence(scene, _playbackType, _skipFrames, _repeatCount);
			var f:String = frame.toString();
			showFrame(parseInt(f));
			_isPlaying = false;
		}
		
		/**
		 * Moves to the next frame of the animation. NOTE: Image Atlas 
		 * animations have more playback options than Flash animations so the
		 * next frame is considered to be the next frame that would be played
		 * under the current play mode so reverse or yo-yo animations may move
		 * the play-head backwards.
		 */		
		override public function nextFrame():void
		{
			if (_anim == null) {
				dispatchEvent(new Event(ANIMATION_FINISHED_EVENT));
				return;
			}
			var animDone:Boolean = false;
			var n:int = _currentFrame + _currentDirection;
			if (n >= _anim.length) {
				n = _anim.length - 1;
				switch (_playbackType) {
					case PLAYBACK_MANUAL:
						animDone = true;
						break;
					case PLAYBACK_FORWARD:
						++_currentRepetition;
						if (_currentRepetition >= _repeatCount) {
							animDone = true;
						} else {
							n = 0;
						}
						break;
					case PLAYBACK_BACKWARD:
						break;
					case PLAYBACK_YOYO:
						_currentDirection = -1;
						break;
				}
			} else if (n <= 0) {
				n = 0;
				switch (_playbackType) {
					case PLAYBACK_MANUAL:
						animDone = true;
						break;
					case PLAYBACK_FORWARD:
						break;
					case PLAYBACK_BACKWARD:
						++_currentRepetition;
						if (_currentRepetition >= _repeatCount) {
							animDone = true;
						} else {
							n = _anim.length - 1;
						}
						break;
					case PLAYBACK_YOYO:
						++_currentRepetition;
						if (_currentRepetition >= _repeatCount) {
							animDone = true;
						} else {
							_currentDirection = 1;
						}
						break;
				}
				
			}
			showFrame(n);
			if (animDone) {
				_isPlaying = false;	
				dispatchEvent(new Event(ANIMATION_FINISHED_EVENT));
			}
		}

		/**
		 * Starts (continues) playing the animation. 
		 */		
		override public function play():void
		{
			showFrame(_currentFrame);
			_isPlaying = true;
		}
		
		/**
		 * Moves to the previous frame of the animation. NOTE: Image Atlas 
		 * animations have more playback options than Flash animations so the
		 * next frame is considered to be the next frame that would be played
		 * under the current play mode so reverse or yo-yo animations may move
		 * the play-head forwards.
		 */		
		override public function prevFrame():void
		{
			if (_anim == null) return;
			var n:int = _currentFrame - _currentDirection;
			if (n < 0)
				n = 0;
			else if (n >= _anim.length)
				n = _anim.length - 1;
			showFrame(n);
		}

		/**
		 * Stops playing the animation. The current frame (and play mode 
		 * information) is retained so the animatoin can be resumed later
		 * by using the play() method.
		 */
		override public function stop():void
		{
			showFrame(_currentFrame);
			_isPlaying = false;
		}

		/**
		 * returns the number of frames in the animation sequence.
		 *  
		 * @return the number of frames in the animation sequence.
		 */
		override public function get totalFrames():int
		{
			if (_anim == null)
				return 1;
			else
				return _anim.length;
		}
		
	}
}