/*
Copyright (C) 2011 by Blazing Games Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com.blazinggames.cgl.atlas
{
	import flash.display.BitmapData;
	import flash.display.Loader;
	import flash.display.LoaderInfo;
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.events.IOErrorEvent;
	import flash.geom.Rectangle;
	import flash.net.URLRequest;

	/**
	 * An image atlas is made up of multipe pages, with each page being an 
	 * image. The image can be loaded externally, or be manually specified if
	 * the images are being embedded.
	 *  
	 * @author Billy D. Spelchan 
	 */	
	public class PageInfo
	extends EventDispatcher
	{
// ***** CONSTANTS *****
		
		public static const PAGE_LOADED_EVENT:String = "pageLoaded";
		
		public static const DEFAULT_ERROR_WIDTH:uint = 256;
		public static const DEFAULT_ERROR_HEIGHT:uint = 256;
		public static const DEFAULT_ERROR_COLOR:uint = 0xFFFF88FF;
		
// ***** VARIABLES *****
		
		private var _pageName:String;
		private var _urlName:String;
		private var _bitmapData:BitmapData;
		private var _errorWidth:uint;
		private var _errorHeight:uint;
		private var _errorColor:uint;
		private var _description:String;
		
		private var _loader:Loader;
		
// ***** CONSTRUCTOR *****
		
		/**
		 * Creates a new page for the image atlas.
		 *  
		 * @param name name of the page.
		 * @param bitmapData the bitmap for the page.
		 */		
		public function PageInfo(name:String = "unknown", bitmapData:BitmapData = null)
		{
			super();
			_pageName = name;
			_urlName = null;
			_bitmapData = bitmapData;
			_errorWidth = DEFAULT_ERROR_WIDTH;
			_errorHeight = DEFAULT_ERROR_HEIGHT;
			_errorColor = DEFAULT_ERROR_COLOR;
			_description = "No description";
		}

// ***** METHODS - XML Handling *****

		/**
		 * Loads the page data from an XML tag. The page tag has the format:
		 * <PAGE name="page name" url="source url" w="width" h="height" color="error color">
		 * description of page</PAGE>
		 * 
		 * @param xml the xml data to parse
		 * @param load should the image be loaded. 
		 */		
		public function setFromXML(xml:XML, load:Boolean = true):void
		{
			_pageName = xml.@name;
			_urlName = xml.@url;
			_bitmapData = null;
			_errorWidth = parseInt(xml.@w);
			_errorHeight = parseInt(xml.@h);
			_errorColor = parseInt(xml.@color);
			_description = xml.toString();
			if (load) {
				loadPage();
			}
		}

		/**
		 * Converts the page into an XML string.
		 *  
		 * @return string representation of page XML tag.
		 */		
		public function toXMLString():String
		{
			return "<PAGE name=\'"+_pageName+
					"\' url=\'"+_urlName+
					"\' w=\'"+_errorWidth+
					"\' h=\'"+_errorHeight+
					"\' color=\'"+_errorColor+
					"'>"+_description+"</PAGE>";
		}
		
// ***** METHODS - Getters and Setters *****
		
		/**
		 * If the page does not load or is null, an error page is created with
		 * the provided parameters.
		 *  
		 * @param w width of error page.
		 * @param h height of error page.
		 * @param c color of error page.
		 * 
		 */
		public function setErrorPage(w:uint, h:uint, c:uint):void
		{
			_errorWidth = w;
			_errorHeight = h;
			_errorColor = c;
		}

		/**
		 * returns the name of the page.
		 *  
		 * @return the name of the page. 
		 */		
		public function get pageName():String
		{
			return _pageName;
		}

		/**
		 * Sets the name of the page
		 *  
		 * @param value name of the page
		 */		
		public function set pageName(value:String):void
		{
			_pageName = value;
		}

		/**
		 * The source url for obtaining the image.
		 * 
		 * @return The source url for obtaining the image.
		 */		
		public function get urlName():String
		{
			return _urlName;
		}

		/**
		 * The source url for obtaining the image. 
		 * 
		 * @param value The source url for obtaining the image.
		 */		
		public function set urlName(value:String):void
		{
			_urlName = value;
		}

		/**
		 * gets the bitmap data for the page
		 * @return BitmapData for the page, null if page not loaded.
		 */
		public function get bitmapData():BitmapData
		{
			return _bitmapData;
		}

		/**
		 * Sets the BitmapData that represents the page. 
		 * 
		 * @param value BitmapData representation of the page.
		 */		
		public function set bitmapData(value:BitmapData):void
		{
			_bitmapData = value;
		}

		/**
		 * Checks to see if the image for the page has been loaded.
		 * 
		 * @return true if page has been loaded otherwise false.
		 */		
		public function get isPageLoaded():Boolean
		{
			return _bitmapData == null;
		}

		/**
		 * returns a text description of the page contents.
		 *  
		 * @return a text description of the page contents.
		 */		
		public function get description():String
		{
			return _description;
		}

		/**
		 * Sets a text description of the page contents. 
		 * 
		 * @param value a text description of the page contents.
		 */		
		public function set description(value:String):void
		{
			_description = value;
		}

// ***** METHODS - Loading *****
		
		/**
		 * Loads the image for this page from the indicated source url. 
		 */		
		public function loadPage():void
		{
			if (_bitmapData != null) { 
				dispatchEvent(new Event(PAGE_LOADED_EVENT));
				return;
			}
			
			if (_loader != null)
				return;
			_loader = new Loader();
			
			var request:URLRequest = new URLRequest(_urlName);
			_loader.contentLoaderInfo.addEventListener(Event.COMPLETE, onLoaded);
			_loader.contentLoaderInfo.addEventListener(IOErrorEvent.IO_ERROR, onIOError);
			_loader.load(request);					
		}
		
		/**
		 * Internal method for converting loaded image into bitmap data when load fails.
		 *  
		 * @param event IO_ERROR event
		 */
		protected function onIOError(event:IOErrorEvent):void
		{
			_bitmapData = new BitmapData(_errorWidth, _errorHeight, true, _errorColor);
			dispatchEvent(new Event(PAGE_LOADED_EVENT));
			trace("page " + _pageName + " failed to loaded.");
		}
		
		/**
		 * Internal method for converting loaded image into bitmap data.
		 *  
		 * @param event COMPLETE event 
		 */
		protected function onLoaded(event:Event):void
		{
			_bitmapData = new BitmapData(_loader.width, _loader.height, true, 0);
			_bitmapData.draw(_loader);
			_loader.unloadAndStop();
			dispatchEvent(new Event(PAGE_LOADED_EVENT));
			trace("page " + _pageName + " loaded.");
		}

// ***** METHODS - Debugging aids *****
		
		override public function toString():String
		{
			return _pageName + " @ " + _urlName + " -> " + _description;
		}
	}
}